/* File: vmain.c */

/*----------------------------------------------------------------------**
** Copyright (C) The Higher Education Statistics Agency 1994,1995,1996  **
** Authors:                                                             **
** 	Duncan Charlton: DuncanC@hesa.ac.uk                                 **
** 	Adrian Crossley: AdrianC@hesa.ac.uk                                 **
** 	John Turner    :   JohnT@hesa.ac.uk                                 **
**----------------------------------------------------------------------*/

/*------------------------------------------------------------------------**
** HESA Validation Program.                                               **
** =======================                                                **
**                                                                        **
** The files that make up this package are:                               **
** 	yynnn.c				Record dependant code                                   **
** 	filehelp.c		CSV and Fixed Length helper routines                    **
** 	helper.c			Misc. Helper routines                                   **
** 	helper.h			declarations from helper.c                              **
** 	nprofile.c		DOS ini-type file functions                             **
** 	vldtn.h				odd declarations                                        **
** 	vmain.c 			Main validation loop                                    **
**  values.c        		array of values                                                               **
** ANSI Compilers - Make sure that #define ANSI appears at the top        **
** 									of vldtn.h                                            **
**                                                                        **
** K&R compilers - 	Make sure that #define ANSI does not appear at        **
** 									the top of vldtn.h                                    **
**                                                                        **
** To view the source files at their best set tabs at positions 3,5,7...  **
**------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "vldtn.h"
#include "helper.h"

long ulConnection=0;
int iCentYear;
int iCurrentYear;
char szlInfoFile[32];
static SZ szMdName[FPSIZE];
static SZ szInputName[FPSIZE];
static SZ szOutputName[FPSIZE];

static SZ szCtlFilename[16];
static SZ szDatFilename[16];
SZ szAuxPath[FPSIZE]="../";

char ubCheckSum=0;

SZ szInst[6];
SZ szRecid[6];
SZ szLineBuffer[LBSIZE];

UINT uiLineLength=0;


UINT uiCurFieldNo=0;
ULONG ulCurRecordNo=0;
UINT uiValFieldResult;
int iActualNoFields,StaffAdjust=0;
/* new error report files */
FILE *pfError1=NULL; 
FILE *pfExplanErr =NULL;
FILE *pfError2=NULL;
FILE *pfError3=NULL;
    

FILE *pfIn=NULL;
FILE *pfError=NULL;
FILE *pfOut=NULL;

ULONG ulIgnoreRules[IRCOUNT];
UINT  uicIgnoreRules=0;
UINT 	uiWarning=10;
UINT 	uiOutput=0;
 
ULONG uinError=0;
ULONG uinWarning=0;
ULONG ulMaxErrors=1000;
ULONG ulMaxWarnings=1000;
SZ szMsgBuffer[512];
BOOL gbLateResultReturn;
unsigned long ulStart;
unsigned long ulEnd;


/*-----------------------------------------------------------**
** 	Default Options are CSV, No Raw Output, Not verbose mode **
**-----------------------------------------------------------*/

/* added an extra 0 on end for bIgnoreFatal - 2004-03-29 (MDE) */

struct tagOptions Options={TY_CSV,0,0,0,{0},0,0,0,0,0,0,0,0};



/*------------------------------**
**    Display Usage Information **
**------------------------------*/
/* dislays list of command line options */
#ifdef ANSI
void Usage(void)
#else
void Usage()
#endif
{

	if (!PIPING) {
		/* Not if we're piping */

		/*fprintf(stderr,"HESA Record Validation Program. Build %s.%s\n",VERMAJ,VERMIN);*/
		fprintf(stderr,"HESA Record Validation Program. \n");
		fprintf(stderr,"Validation Rules Version %s\n",szVRVer);
		fprintf(stderr,"Date written: %s\n",szDateWritten);  	
		fprintf(stderr,"Copyright (C) The Higher Education Statistics Agency 2005\n");
		fprintf(stderr,"Usage: %s <filename> [switches]\n\n",szMdName);
		fprintf(stderr,"Switches are:\n");	                                                    
		/*	           1         2         3         4         5         6         7          */
		/*	 01234567890123456789012345678901234567890123456789012345678901234567890123456789 */
		
		/* 2004-02-19 (ME) amended to display switches available in standalone mode: */
		/* 2006-02-21  (ME) added -uc, -uf */

		fprintf(stderr,"\t-b\tWait for keypress before exiting\n");
		fprintf(stderr,"\t-e<n>\tMaximum number of errors (Default 1000)\n");
		fprintf(stderr,"\t-g<n>\tCountry of Institution : England(g1), Scotland(g2),\n\t\t Wales(g3), N.Ireland(g4)\n");
		fprintf(stderr,"\t-i<nnn>\tDon't validate field <nnn>\n");
		fprintf(stderr,"\t-j<nnn>\tIgnore rule <nnn>\n");
		fprintf(stderr,"\t-n\tIgnore all fields EXCEPT those given by -i\n");
		fprintf(stderr,"\t-p<dir>\tlook for jacscode,postcode in <dir>\n");
		fprintf(stderr,"\t-q\tQuiet mode\n");
		fprintf(stderr,"\t-h\tError report in database input format\n");
		fprintf(stderr,"\t-r\tRaw Output\n");
		fprintf(stderr,"\t-tc\tInput is comma separated\n");
		fprintf(stderr,"\t-tf\tInput is fixed length\n");
		fprintf(stderr,"\t-uc\tOutput in comma separated format, including full English text\n");
		fprintf(stderr,"\t-uf\tOutput in fixed length format, including full English text\n");
		fprintf(stderr,"\t-v\tVerbose Output\n");
		fprintf(stderr,"\t-w<nn>\tSet warning level (-w0) to include warnings in error file\n");
		fprintf(stderr,"\t-x\tExplain Errors (Wide Output)\n");
		fprintf(stderr,"\t-y\tAllows Data Structure Errors in data\n");
		
		exit(-1);
		}
		
}

/*
 * Report Internal Error and exit
 */

#ifdef ANSI
void ErrExit(UINT uiError)
#else
void ErrExit(uiError)
UINT uiError;
#endif
{
/* char ErrNo[8];   commented out, is not used anywhere, MDE 2005-05-07 */
char ErrorStr[7][14]= {"IERR_USAGE","IERR_MINPUT","IERR_SWITCH","IERR_FINPUT","IERR_FOUTPUT","IERR_MEMORY"};

  if(!PIPING){

		if (uiError!=IERR_USAGE)                                           /* IERR_USAGE, when no arguments*/
  			fprintf(stderr,"%s- Internal Error: #%u\n",szMdName,uiError);

		/* FileWritePrivateProfileString("Flags","InternalError",ErrorStr[uiError-1],szlInfoFile); */
		fclose(stderr);

		if (Options.bBreak)
	  		getchar();
    	}

  else {
		fprintf(stderr,"%08lu #############    ### ######## %u      ####   #\n",ulCurRecordNo,uiError);
		#ifdef ANSI
		#else
			itoa(uiError,ErrNo,10);
		#endif
				
		FileWritePrivateProfileString("Flags","InternalError",ErrorStr[uiError-1],szlInfoFile);
	 	exit(uiError);
		}
	/* changed to return -1 for Ardvarrk*/
	exit(-1);
	FileWritePrivateProfileString("Main","ExitCode","-1",szlInfoFile);
}

/*
 * Report a data structure Error
 */

#ifdef ANSI
void DataError(UINT uiErr,const PSZ pszError)
#else
void DataError(uiErr,pszError)
UINT uiErr;
const PSZ pszError;
#endif
{
	/* ~ strcpy(szInputName,strrchr(szInputName,'\\')+1);    */
    if(!Options.bPipe){
	  if (!uiErr) {
		fprintf(pfError,"\n%-16s Line %08lu: Data Structure Error %s\n",szInputName,ulCurRecordNo,pszError);
		fprintf(stderr,"\n%-16s Line %08lu: Data Structure Error, please see error file. \n",szInputName,ulCurRecordNo,pszError);
		FileWritePrivateProfileString("Flags","DataStructureError",pszError,szlInfoFile);
		}
	  else 
	    {
		fprintf(pfError,"\n%-16s Line %08lu: Data Structure Error %s\n",szInputName,ulCurRecordNo,pszError);
		fprintf(stderr,"\n%-16s Line %08lu: Error in Field %03u %s\n",szInputName,ulCurRecordNo,uiCurFieldNo,pszError);
		}

	  if (Options.bBreak)
		getchar();
		   /*changed to -1 	 for Ardvarrk to get a 1 SUccess 0 Fail -1 Error*/

	  if(!Options.bIgnoreFatal)   /* added in 2004-03-29 (MDE) */
	    exit(uiErr);         
	}
	else{
	  fprintf(pfError,"%08lu #############    ### ######## %u     ####   #\n",ulCurRecordNo,uiErr);
	    fprintf(pfError2,"\n%s\n",pszError); /* transno.ERRORS file */
		fprintf(pfError1,"\n%s\n",pszError); /* transno.SUMMARY file */
		FileWritePrivateProfileString("Flags","DataStructureError",pszError,szlInfoFile);
		FileWritePrivateProfileString("Main","ExitCode","0",szlInfoFile);
 	  exit(0);
	}

}
/* gets switches does some error checking eg if a field is asked to be           
excluded from testing ( -i) and it doesnt exist eg 300.*/
void DoArgs(int argc,char **argv) 
{
 	int cArg=1;
  UINT uiTemp=0;

	if (argc==1) {
		Usage();
		ErrExit(IERR_USAGE); /* No args at all */
		}


	while (cArg<argc) {
		/* Loop through the args */
		if (*(argv[cArg])=='-') {
			/* This one must be a switch */
			switch(*(argv[cArg]+1)) {
				/* Max Numbers of Errors to report */


				case 'A':
				case 'a':
					REALLOCUNIT=atoi(argv[cArg]+2);					
					break;

				case 'B':
				case 'b':
					Options.bBreak=1;
					break;

				
				case 'C':
				case 'c':
					Options.bControl=1;
					break;
	
				case 'D':
				case 'd':						
					Options.uiDiag=atoi(argv[cArg]+2);					
					break;				
				
				case 'E':
				case 'e':			              
					ulMaxErrors=atol(argv[cArg]+2);
					ulMaxWarnings =ulMaxErrors;
					break;

				case 'F':
				case 'f':
					Options.bPipe=1;
					ulConnection=atol(argv[cArg]+2);
					break;
			
				/* Tell us which global switches */
				case 'G':
				case 'g':
				case 'L':
				case 'l':
					switch(*(argv[cArg]+2)) {
					case '1': 
						Options.bGlobal[1]=1;
						break;
					case '2': 
						Options.bGlobal[2]=1;
						break;
					case '3':
						Options.bGlobal[3]=1;
						break; 
					case '4':          
						Options.bGlobal[4]=1;												
						break; 
					case '5':
						Options.bGlobal[5]=1;
						break;
					case '6':
						Options.bGlobal[6]=1;
						*(szErrorFile+2)='1';
						*(szlInfoFile+2)='1';
						break;
					default: 
						if (!PIPING)
							fprintf(stderr,"Unknown command line switch: %s\n",argv[cArg]);
						ErrExit(IERR_SWITCH); /* Bad Switch */
						}
					break;
				case 'h':
				case 'H': 
				    Options.bDBOutput = 1;
				    break;	  
	  
				/* We want to ignore a field */					  
				case 'I':
				case 'i':
					uiTemp=atoi(argv[cArg]+2);
					
					if (uiTemp==0|| uiTemp>uinFields) {
						if (!PIPING)
							fprintf(stderr,"Field number out of range %u\n",uiTemp);
						ErrExit(IERR_SWITCH); /* Field No is out of range */
						}
											
					stFields[uiTemp-1].bIgnore=1;
					                                                       
					break;
					
				/* We want to ignore a rule */					                             
				case 'J':
				case 'j':					 
				
				  if (uicIgnoreRules>=IRCOUNT) {                            
						if (!PIPING)
				  		fprintf(stderr,"Too many rules to ignore.");
				  	ErrExit(IERR_SWITCH); /* Too many ignored rules */
				  	}
				
					ulIgnoreRules[uicIgnoreRules++]=atoi(argv[cArg]+2);
					break;                 

				/* Additional path for data files */
        
				case 'N':
				case 'n':	
					Options.bNegate=1;					
					break;        
        
				case 'O':
				case 'o':	
					strcpy(szOutputName,argv[cArg]+2);
					break;        
        
				case 'P':
				case 'p':	
					strcpy(szAuxPath,argv[cArg]+2);
					break;
					
				case 'Q':
				case 'q':	
					Options.bQuiet=1;										
					break;
					
				/* Asking for RawOutput */
				case 'R':
				case 'r':
					Options.bRawOutput=1;
					break;

				case 'S':
				case 's':	
					Options.uiReduced=atoi(argv[cArg]+2);
					break;

									
				/* Tell us the type of file */
				case 'T':
				case 't':
					switch(*(argv[cArg]+2)) {
					case 'C': /* CSV */
					case 'c':
						Options.uiType=TY_CSV;
						break;
					case 'F': /* Fixed */
					case 'f':
						Options.uiType=TY_FIX;
						break;
					case 'T':/* Separated with tabs */
					case 't':
						Options.uiType=TY_TSV;
						break;
					default:
						if (!PIPING)
							fprintf(stderr,"Unknown command line switch: %s\n",argv[cArg]);
						ErrExit(IERR_SWITCH); /* Bad Switch */
						}
					break;

				/* Output in csv or fixed format to include full english text MDE 2006-02-21*/
				/* see also overrides at end of this function */
				case 'U':
				case 'u':
					switch(*(argv[cArg]+2)) {
					case 'C': /* CSV */
					case 'c':
						Options.uoType=TY_CSV;
						break;
					case 'F': /* Fixed */
					case 'f':
						Options.uoType=TY_FIX;
						break;
					case 'T':/* Separated with tabs */
					case 't':
						Options.uoType=TY_TSV;
						break;
					default:
						if (!PIPING)
							fprintf(stderr,"Unknown command line switch: %s\n",argv[cArg]);
						ErrExit(IERR_SWITCH); /* Bad Switch */
						}
					break;

				case 'V':
				case 'v':
					Options.bVerbose=1;
					break;             
					
				case 'W':
				case 'w':
					uiWarning=atoi(argv[cArg]+2);					
					break;					

				case 'X':
				case 'x':
					Options.bExplain=1;					
					break;					

				case 'Y':
				case 'y':
					Options.bIgnoreFatal = 1;
					break;	

				case 'Z':
				case 'z':
					Options.bNoValid=1;					
					break;					

			default:
				if (!PIPING)
					fprintf(stderr,"Unknown command line switch: %s\n",argv[cArg]);
				ErrExit(IERR_SWITCH); /* Bad Switch */
				
				}
			}
		else {
			/* Not a switch, must be the input filename */
			if (*szInputName) {
				if (!PIPING)
					fprintf(stderr,"Too many input files: %s\n",argv[cArg]);				
				ErrExit(IERR_MINPUT); /* Can only specify one input file */
				}
			/* Store the file name */
			strcpy(szInputName,argv[cArg]);
    		if(strrchr(szInputName,'\\'))
				strcpy(szInstFromFileName,strrchr(szInputName,'\\')+1);
				 else if(strrchr(szInputName,'/'))
				      strcpy(szInstFromFileName,strrchr(szInputName,'/')+1);   
				else
				strcpy(szInstFromFileName,szInputName);
			szInstFromFileName[4]='\0';                             

			}

		cArg++;
		}

		/* -uc and -uf are to cancel -h and -x */
		if (Options.uoType==TY_CSV||Options.uoType==TY_FIX){
				Options.bDBOutput=0;
				Options.bExplain=0;
				Options.bRawOutput=0;
				Options.bVerbose=0;
		}


}
/* piping options are initilised.*/
void SetPipingOptions(void)
{
	Options.bVerbose=0;
	Options.bRawOutput=1;
	Options.bExplain=1;
	Options.bControl=0;
	Options.bBreak=0;
	Options.bNoValid=0;
	Options.uiDiag=0;
	/*Options.bNegate=0;    COMMENTED OUT BY MDE 2003-05-16 */
	Options.bQuiet=1;
	Options.bIgnoreFatal=0;  /* added in 2003-03-29 (ME) */
	Options.uoType=TY_FIX;  /* added in 2006-02-21 (ME) */
	
}
                                     
/* The total and summary error files are filled up.*/
void DoError1and4Report(struct ERROR *pstdErrors,struct FLD *pfld)                                        
{                           /* ~ do first report first   */
	static char ErrorText[256];

	static char Line1[]=    {"Field : "};
	static char Warning[] = {" (Warning) : "};
	static char Error[] =   {" ( Error ) : "};
	static char Err_Warn[15];   
	static char Count_Of_Rec[]= {"Count of Records = "};
	static char Rule[]={"Rule "};	
    static char Underline[]={"====================\n"}; 
    unsigned int ErrorTableIndex=0;
	unsigned int CurrentField ;
	if(!uinError){
		/* .SUMMARY */
		fprintf(pfError1,"No errors in data ");		 
		/* .ERRORS */
		fprintf(pfError2,"No errors in data");
		};
    if(!uinWarning)
		fprintf(pfError3,"No Warnings in data");
		
	while(ErrorTableIndex<uinRules){
		if(((pstdErrors+ErrorTableIndex)->uiErrorTotal)!=0){          /* ~ this lot is the header for each field */
			fprintf(pfError1,"\n\n%s%03u %s",Line1,(pstdErrors+ErrorTableIndex)->uiFieldNo,(pfld+(pstdErrors+ErrorTableIndex)->uiFieldNo-1)->szAbbrev);
   			fprintf(pfError1,"\n%s",Underline);                   
	    	CurrentField = (pstdErrors+ErrorTableIndex)->uiFieldNo;                                           
	    	                                                      /* ~ this loop does  errors for each field  */
    	    while((pstdErrors+ErrorTableIndex)->uiFieldNo== CurrentField){
    	      	if(((pstdErrors+ErrorTableIndex)->uiErrorTotal)!=0){
					if((pstdErrors+ErrorTableIndex)->uiSeverity<uiWarning)
						strcpy(Err_Warn,Warning);
	   	     		else
	           			strcpy(Err_Warn,Error);
	        		fprintf(pfError1,"\n%s%04d%s%s%u",Rule,(pstdErrors+ErrorTableIndex)->uiErrorNumber,Err_Warn,Count_Of_Rec,(pstdErrors+ErrorTableIndex)->uiErrorTotal);
					strcpy(ErrorText,GetLongErrorMessage((pstdErrors+ErrorTableIndex)->uiErrorNumber));
					/* ~ added 27/11/00  */
					fprintf(pfError1,"\n%s\n",ErrorText);
					ErrorTableIndex++;	    	                        
				    }   
				else
				    ErrorTableIndex++;	 
 		    	}
		  }                       
		  else
		  	ErrorTableIndex++;
		}	
}	     
/*attempts to open and read into memory short error file,postcode,hesacode and valid entries*/                          
void ReadDataFiles(void)
{
	if (Options.bExplain)  {
		/* Won't happen if piping */
		if ((pfErrors=fopen(szBinFile,"rb"))==NULL) {
			fprintf(stderr,"Can't open: %s\n",szBinFile);  
			ErrExit(IERR_FINPUT); /* Can't open that */
			}

		ReadErrorTable();
		}

	/* Try to read the POSTCODE file */                     
	if (uiNeedsPostcode)
		ReadPostCODE();
	
	/* Try to read the JACSCODE file */                     
	if (uiNeedsJACScode)
		ReadJACSCODE();
	
    /* Read FE Colleges */
    if(uiRNo==12)
    ReadFEColleges();

	/* Try to read the valid entries file */                     
	ReadValidEntries();

}

/*does header for error report.*/
void DoHeader(void) 
{
	if (!PIPING) {
		/* standalone mode. Header is not required when producing raw output  MDE 2004-05-19 */
		if (!Options.bRawOutput) {
	

			if( uiRNo==11 ||uiRNo==12) {
			
   				if ((Options.uoType == TY_FIX)||(Options.uoType == TY_CSV)){
					if (Options.uoType == TY_FIX) {            
							fprintf(pfError,"%-8s %-16s %-20s %-20s %-3s %-8s %-6s %-4s %-46s %-100s\n","Record","ID","Ownstu","OwnPsd","Fld","Abbrev","Error","Sev."," Data","English Text of Rule");
					}
					else {
							fprintf(pfError,"%-s,%-s,%-s,%-s,%-s,%-s,%-s,%-s,%-s,%-s\n","Record","ID","Ownstu","OwnPsd","Fld","Abbrev","Error","Sev.","Data","English Text of Rule");
					}
				}
				else
					{	
						/* student (type 11 and 12) records */
						if (!Options.bExplain && !Options.bDBOutput) 		
						fprintf(pfError,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s  %s %16s %16s\n","Record","ID","Fld","Abbrev","Error","Sev.","Data","Ownstu","OwnPsd");
						
						if(Options.bExplain && !Options.bDBOutput) 
							fprintf(pfError,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s %-32s  %s  %16s %16s\n","Record","ID","Fld","Abbrev","Error","Sev.","Error Text","Data","Ownstu","OwnPsd");
						       
						if (!Options.bExplain && Options.bDBOutput) 		
							fprintf(pfError,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s %100s %16s %16s\n","Record","ID","Fld","Abbrev","Error","Sev.","Data","Ownstu","OwnPsd");
						
						if(Options.bExplain && Options.bDBOutput)
							fprintf(pfError,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s %-32s %100s  %16s %16s\n","Record","ID","Fld","Abbrev","Error","Sev.","Error Text","Data","Ownstu","OwnPsd");

					}

			}
			else {

				/* not 11 or 12 so do not output ownstu and ownpsd */

   				if ((Options.uoType == TY_FIX)||(Options.uoType == TY_CSV)){
                /* it is format -uf or -uc */
					if (Options.uoType == TY_FIX) {
                        if (uiRNo==26) {
		                    fprintf(pfError,"%-8s %-16s %-20s %-3s %-8s %-6s %-4s %-40s %-500s\n","Record","ID","Contract ID","Fld","Abbrev","Error","Sev.","Data","English Text of Rule");
                        }
                        else {
							fprintf(pfError,"%-8s %-16s %-3s %-8s %-6s %-4s %-40s %-500s\n","Record","ID","Fld","Abbrev","Error","Sev.","Data","English Text of Rule");
                        }
					}
					else {
                        if (uiRNo==26)  {
							fprintf(pfError,"%-s,%-s,%-s,%-s,%-s,%-s,%-s,%-s,%-s\n","Record","ID","Contract ID","Fld","Abbrev","Error","Sev.","Data","English Text of Rule");
                        }
                        else {
                            fprintf(pfError,"%-s,%-s,%-s,%-s,%-s,%-s,%-s,%-s\n","Record","ID","Fld","Abbrev","Error","Sev.","Data","English Text of Rule");
                        }
					}
				}
				else {


			/* formats other than -uc or -uf */
					if (!Options.bExplain && !Options.bDBOutput) 		
						fprintf(pfError,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s  %s\n","Record","ID","Fld","Abbrev","Error","Sev.","Data");
					
					if(Options.bExplain && !Options.bDBOutput) 
						fprintf(pfError,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s %-32s  %s\n","Record","ID","Fld","Abbrev","Error","Sev.","Error Text","Data");
					       
					if (!Options.bExplain && Options.bDBOutput) 		
						fprintf(pfError,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s %100s\n","Record","ID","Fld","Abbrev","Error","Sev.","Data");
					
					if(Options.bExplain && Options.bDBOutput)
						fprintf(pfError,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s %-32s %100s\n","Record","ID","Fld","Abbrev","Error","Sev.","Error Text","Data");
				}		         
			}
	   }
	}
	else {
		/* aardvark mode: fixed headers for .ERROR and .WARNING files */
         if (uiRNo==26)  {
			fprintf(pfError2,"\n\n%-8s %-16s %-20s %-3s %-8s %-6s %-4s %-32s %s\n","Record","ID","Contract ID","Fld","Abbrev","Rule","Sev.","Error Text","Data");
			fprintf(pfError3,"\n\n%-8s %-16s %-20s %-3s %-8s %-6s %-4s %-32s %s\n","Record","ID","Contract ID","Fld","Abbrev","Rule","Sev.","Error Text","Data");					
         }
         else {
   			fprintf(pfError2,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s %-32s %s\n","Record","ID","Fld","Abbrev","Rule","Sev.","Error Text","Data");
			fprintf(pfError3,"\n\n%-8s %-16s %-3s %-8s %-6s %-4s %-32s %s\n","Record","ID","Fld","Abbrev","Rule","Sev.","Error Text","Data");					
         }
    }	   		
}

/* updates .inf file with stuff like no. Of errors warnings */
void UpDateInfoFile(int argc,char **argv)
{
	int uiTemp=0;    
	static int iFlags=0;
	char *pszScan=NULL;

	
	if (!PIPING) {	
		fprintf(stdout,"\n\nComplete. %lu lines from %s were processed \n",ulCurRecordNo,szInputName);
		fprintf(stdout,"%lu seconds elapsed. %lu records/second.\n",ulEnd-ulStart,(ulEnd-ulStart)?ulCurRecordNo/(ulEnd-ulStart):ulCurRecordNo);	
		fprintf(stdout,"%lu Errors in %s\n",uinError,szErrorFile);
		fprintf(stdout,"%lu Warnings in %s\n",uinWarning,szErrorFile);
		}

                                                     
	sprintf(szMsgBuffer,"%lu",ulCurRecordNo);
	FileWritePrivateProfileString("Main","NumberOfRecords",szMsgBuffer,szlInfoFile);
	sprintf(szMsgBuffer,"%lu",uinError);                           
	FileWritePrivateProfileString("Main","NumberOfErrors",szMsgBuffer,szlInfoFile);
	sprintf(szMsgBuffer,"%lu",uinWarning);                           
	FileWritePrivateProfileString("Main","NumberOfWarnings",szMsgBuffer,szlInfoFile);

	sprintf(szMsgBuffer,"%u",ubCheckSum);
	FileWritePrivateProfileString("Main","Checksum",szMsgBuffer,szlInfoFile);		                                                                   
	sprintf(szMsgBuffer,"%lu",ulEnd-ulStart);
	FileWritePrivateProfileString("Time","SecondsElapsed",szMsgBuffer,szlInfoFile);		                                                                   
	sprintf(szMsgBuffer,"%lu",(ulEnd-ulStart)?ulCurRecordNo/(ulEnd-ulStart):ulCurRecordNo);
	 FileWritePrivateProfileString("Time","SecondsPerRecord",szMsgBuffer,szlInfoFile);		                                                                   
    
	for (uiTemp=0;uiTemp<argc;uiTemp++) {
		sprintf(szMsgBuffer,"Arg%03u",uiTemp);                     
		FileWritePrivateProfileString("Args",szMsgBuffer,argv[uiTemp],szlInfoFile);
		}
  
	sprintf(szMsgBuffer,"%u",argc);                     	  
		FileWritePrivateProfileString("Args","nArgs",szMsgBuffer,szlInfoFile);
	
	if ( (Options.bControl) && (PIPING) ) {

		sprintf(szCtlFilename,"%s_%03u.ctl",szInst,uiRNo);   
	  	

		FileWritePrivateProfileString("Institution","HESAID",szInst,szCtlFilename);     

		sprintf(szMsgBuffer,"%s",szLRNo);                     
		FileWritePrivateProfileString("Data","RecordType",szMsgBuffer,szCtlFilename);     


		switch (Options.uiType) {
		case TY_CSV:
			sprintf(szMsgBuffer,"csv");
			break;
		case TY_FIX:                      
			sprintf(szMsgBuffer,"asc");				
			break;
			}
  
		fprintf(stdout,"\nUpdating Control File %s...\n",szCtlFilename);    
  
		FileWritePrivateProfileString("Data","Format",szMsgBuffer,szCtlFilename);     
		fprintf(stdout,"    Format=%s\n",szMsgBuffer);
                                                              

		sprintf(szMsgBuffer,"%lu",ulCurRecordNo);
		FileWritePrivateProfileString("Control","NumberOfRecords",szMsgBuffer,szCtlFilename);
		fprintf(stdout,"    NumberOfRecords=%s\n",szMsgBuffer);

		sprintf(szMsgBuffer,"%u",ubCheckSum);
		FileWritePrivateProfileString("Control","Checksum",szMsgBuffer,szCtlFilename);
		fprintf(stdout,"    Checksum=%s\n",szMsgBuffer);
		FileWritePrivateProfileString("Control","AutoControl","True",szCtlFilename);
	                                                                   

/*
		FileGetPrivateProfileString("Control","Version","0",szMsgBuffer,32,szCtlFilename);
	  sprintf(szMsgBuffer,"%u",atoi(szMsgBuffer)+1);	                                     
		FileWritePrivateProfileString("Control","Version",szMsgBuffer,szCtlFilename);			  
*/		
		
		FileGetPrivateProfileString("Control","Transaction","",szMsgBuffer,32,szCtlFilename);

		if (!*szMsgBuffer) {
			FileWritePrivateProfileString("Control","Transaction","Create",szCtlFilename);			  
			fprintf(stdout,"    Transaction=%s\n","Creat");
			}


		FileGetPrivateProfileString("Contact","Name","",szMsgBuffer,32,szCtlFilename);
		if (!*szMsgBuffer) {
			FileWritePrivateProfileString("Contact","Name","",szCtlFilename);			  
			fprintf(stdout,"    Name=%s\n","");
			iFlags|=0x01;                    
			}

		FileGetPrivateProfileString("Contact","Email","",szMsgBuffer,32,szCtlFilename);
		if (!*szMsgBuffer) {
			FileWritePrivateProfileString("Contact","Email","",szCtlFilename);			  
			fprintf(stdout,"    Email=%s\n","");
			iFlags|=0x02;
			}           

		FileGetPrivateProfileString("Contact","Telephone","",szMsgBuffer,32,szCtlFilename);
		if (!*szMsgBuffer) {
			FileWritePrivateProfileString("Contact","Telephone","",szCtlFilename);			  
			fprintf(stdout,"    Telephone=%s\n","");
			iFlags|=0x04;
			}

		FileGetPrivateProfileString("Contact","Fax","",szMsgBuffer,32,szCtlFilename);
		if (!*szMsgBuffer) {
			FileWritePrivateProfileString("Contact","Fax","",szCtlFilename);			  
			fprintf(stdout,"    Fax=%s\n","");
			iFlags|=0x08;
			}		
			
		}

     if(PIPING){
	 	if(uinError==0)
	 		FileWritePrivateProfileString("Main","ExitCode","1",szlInfoFile);
		if(uinError>0)
	 		FileWritePrivateProfileString("Main","ExitCode","0",szlInfoFile);
	 }
                  
	if (!PIPING) {

		if ((pszScan=strrchr(szInputName,'/'))==NULL)
			pszScan=strrchr(szInputName,'\\');	                                                              

	
		sprintf(szDatFilename,"%s_%03u.%s",szInst,uiRNo,Options.uiType==TY_CSV?"csv":"asc");   		

	/* 2004-02-03(MDE) commented out:
		if (strcmpi(!pszScan?szInputName:pszScan+1,szDatFilename)) {
			iFlags|=0x10;
			}
	*/

		                    
		if (iFlags)		                    
			fprintf(stdout,"\nNote - the following action(s) need to be taken.\n");      
		                    
		if (iFlags & 0x01) 
			fprintf(stdout,"    Edit the control file to include a Contact Name.\n");      
		
		if (iFlags & 0x04) 		
			fprintf(stdout,"    Edit the control file to include a Contact Telephone #.\n");      
    	  
		if ((iFlags & 0x02) && (iFlags & 0x08)) 
			fprintf(stdout,"    Edit the control file to include either a Contact Email or Fax #.\n");
                        
		if (iFlags & 0x10) 	 
		   fprintf(stdout,"    The Data File should be called %s.\n",szDatFilename);
		
		}

}
/*opens up data (in and out)  and error files*/
void OpenFiles(void) {
	
	
	char szlTotalsName[32];
	char szlError1File[32];
	char szlWarningFile[32];
	char szlLongDescFile[32];

	if (!PIPING) {

		/* validation kit is being run in standalone mode. */

  		fprintf(stdout,"Processing data file %s.\n",szInputName);
    
		/* open the input data file */                            
		if ((pfIn=fopen(szInputName,"r"))==NULL) {
			fprintf(stderr,"Can't open: %s\n",szInputName);
			ErrExit(IERR_FINPUT); /* Can't open that */
			}

		/* open the .err file */
		if ((pfError=fopen(szErrorFile,"w"))==NULL) {
			fprintf(stderr,"Can't open: %s\n",szErrorFile);
			ErrExit(IERR_FOUTPUT); /* Can't open that */
			}

		/* prepare variables for opening the .inf file */
			strcat(szlInfoFile,szLRNo);
			strcat(szlInfoFile,".INF");
		}

	else {

		/* validation kit is being run in aardvark mode */

		sprintf(szlTotalsName,"%08lu",ulConnection);
		strcpy(szlError1File ,szlTotalsName );
		strcpy(szlWarningFile,szlTotalsName ) ;
		strcpy(szlLongDescFile,szlTotalsName );
		strcpy(szlInfoFile,szlTotalsName );
		strcat(szlTotalsName,szTotalsFile);
		strcat(szlError1File,szError1File);
		strcat(szlWarningFile,szWarningFile);
		strcat(szlLongDescFile,szLongDescFile);		
		strcat(szlInfoFile,szInfoFile);

		/* open the input data file. */
		if ((pfIn=fopen(szInputName,"r"))==NULL) {
			ErrExit(IERR_FINPUT); /* Can't open that */
			}
		/* open the output data file as defined by the -o.... switch */
		/* usually <transno>.DATA  */
		pfError=stderr;
		if ((pfOut=fopen(szOutputName,"w"))==NULL) {
			ErrExit(IERR_FINPUT); /* Can't open that */
			}
		   
	    /* open the <recid>rl.csv file */
		if ((pfExplanErr=fopen(szExplanErrFile,"r"))==NULL) {   
			fprintf(stderr,"Can't open: %s\n",szExplanErrFile);	
			ErrExit(IERR_FINPUT); /* Can't open that */
			}   

		/* open the <transno>.SUMMARY file */
		if ((pfError1=fopen(szlTotalsName,"w"))==NULL) {
			fprintf(stderr,"Can't open: %s\n",szTotalsFile);
			ErrExit(IERR_FINPUT); /* Can't open that */
			}   

		/* open the <transno>.ERRORS file */
		if ((pfError2=fopen(szlError1File,"w"))==NULL) {
			fprintf(stderr,"Can't open: %s\n",szError1File);
			ErrExit(IERR_FINPUT); /* Can't open that */
			}                      

		/* open the <transno>.WARNINGS file */
		if ((pfError3=fopen(szlWarningFile,"w"))==NULL) {
			fprintf(stderr,"Can't open: %s\n",szWarningFile);
			ErrExit(IERR_FINPUT); /* Can't open that */
			}	                   
	} 
}

/*current year obtained, argumnts processed, header put out , 
jacscode postcode valid entries read into memory
datafile error fiels opened
dataline read and checked for good data structure
Mixing of Late return of results (LRR ) with normal (0 in 3rd place) checked
Each field in each line is checked by calling the appropiate function that contains all the previously written rules held in a database and exported to a c file.
Main function returns a -1 if there is a program running error or 0 if the program runs but fails the data it is testing or 1 if it passes the data ( A pass or fail is designated as no errrors or some errors).
*/
#ifdef ANSI
void main(unsigned int argc,char **argv)
#else
void main(argc,argv)
unsigned int argc;
char **argv;
#endif
{
  /* ~  int p;   */
  ULONG uinErrorStart=0;
 unsigned long YearSecs=365*86400;
  static char *pszEOL=NULL;
	static int iC;    
  char *pszScan=NULL;
 
  /* ~ pstErrors =&Error_Object[0];   */
  ulStart=time(NULL);            
  iCurrentYear=(int) (ulStart/YearSecs)+1970;
  iCentYear= (int) (iCurrentYear/100)*100;
  
	/* Make a copy of the program name */
	strcpy(szMdName,argv[0]);
	
    
	/* Process all the args */
	DoArgs(argc,argv);
    
  if (!PIPING) {


		/* Say Hello */
		fprintf(stderr,"HESA Record Validation Program\n");
		fprintf(stderr,"%s.\n",szMdName);
		fprintf(stderr,"Validation Rules Version %s\n",szVRVer);
		fprintf(stderr,"Date written: %s\n",szDateWritten);  	
		/* ~ remove(szInfoFile);  */
		sprintf(szMsgBuffer,"Sizeof FLD:%u\n",sizeof(struct FLD));                     
		ShowDiag(szMsgBuffer,1);
		}
  else  {
		SetPipingOptions();
		}
 

	/* Open Files */
	OpenFiles();
	remove(szlInfoFile);
	/* Read in data files */
	ReadDataFiles();

	/* Error Header */
	DoHeader();

	if (!PIPING) {
		sprintf(szMsgBuffer,"%s.%s",VERMAJ,VERMIN);                     
		FileWritePrivateProfileString("Main","Version",szMsgBuffer,szlInfoFile);

		}


    /* MDE 2004-02-03 - moved this outside above !PIPING section so it applies to Aardvark too*/
    sprintf(szMsgBuffer,"%s",szVRVer);                     
	FileWritePrivateProfileString("Main","RulesVersion",szMsgBuffer,szlInfoFile);

	    /* MDE 2004-02-03 added to record the date <recid>.c was built in the INF file*/
    sprintf(szMsgBuffer,"%s",szDateWritten);                     
	FileWritePrivateProfileString("Main","DateWritten",szMsgBuffer,szlInfoFile);

	/* Read the input file a line at a time */
	ulCurRecordNo=1;

	for (iC=0;iC<LBSIZE;iC++) {
		*(szLineBuffer+iC)=0;
		}

	/*get a line while the maxerrors is not reached or maxerrors isn't zero*/
	while (fgets(szLineBuffer,LBSIZE-1,pfIn)!=NULL && (uinError<ulMaxErrors || !ulMaxErrors)) {
		
		            
		uinErrorStart=uinError;
    
    uiLineLength=strlen(szLineBuffer);		                                        
    
		if (uiLineLength) 
			*(szLineBuffer+--uiLineLength)=0; /* zot the trailing '/n' */
                     
		if (uiLineLength) {    
			pszEOL=szLineBuffer+uiLineLength-1;
			while (*pszEOL=='\n' || *pszEOL=='\r' ) {
				*pszEOL=0;
				pszEOL--;
				}
			
			} 
                     
		pszScan=szLineBuffer;
		
		while (*pszScan) {
			if (*pszScan<32 || *pszScan>140) {                                          
				sprintf(szMsgBuffer,"\nIllegal character (%c) in Data File, Line %lu at position %d.\nProcessing terminated due to illegal character.\n",*pszScan,ulCurRecordNo,pszScan-szLineBuffer+1);
				DataError(0,szMsgBuffer);
				}                 
			if (!isspace(*pszScan))				
				ubCheckSum^=*pszScan;				
			pszScan++;				
			}
		if(strstr(szLRNo,"021"))
		     StaffAdjust=1; 	 
		strcpy(szRecid,GetFieldNoString(0));                                        
		if (uiLineLength) {
			switch (Options.uiType) {
			case TY_CSV:
				/* code amended 13/10/95 Adrian Crossley                               */
				/* to return data structure errors on double quotes error in fields    */ 
				iActualNoFields=GetLineOffsetsCSV(stFields,szLineBuffer);
				if (iActualNoFields==-1)
					{
					sprintf(szMsgBuffer,"Illegal character (\") in Data File\n");
					DataError(0,szMsgBuffer);
				    }    
				if (iActualNoFields+StaffAdjust<(int)uinFields && szRecid[2]== '0')
					{
					sprintf(szMsgBuffer,"Only %d fields in Line %08lu\n",iActualNoFields,ulCurRecordNo);
					DataError(0,szMsgBuffer);
				    }        

				break;
			case TY_TSV:
				GetLineOffsetsTSV(szLineBuffer);
				break;
			case TY_FIX:                      
				iActualNoFields=GetLineOffsetsFIX(szLineBuffer);
				if(iActualNoFields+StaffAdjust<(int)uinFields && szRecid[2]== '0'){
					sprintf(szMsgBuffer,"Only %d full field(s) in Line %08lu\n",iActualNoFields,ulCurRecordNo);
					DataError(0,szMsgBuffer);	                                       
					}

				break;
				}
			}                
		else {
			sprintf(szMsgBuffer,"Zero length line(ln= %08lu) in Data File.",ulCurRecordNo);
			DataError(0,szMsgBuffer);
			}			                                    
			      

		if (ulCurRecordNo==1)	{
			strcpy(szInst,GetFieldNoString(1));
			/*  MUST UPDATE */ 
			/* Code added to provide ability to prevent Late Result Return from being mixed*/
			/* with other returns in the same file                                         */
			/* Added 26/04/96 by Adrian Crossley                                           */

			/* NOTE - PIPING is true when the kit is run in Aardvark mode. */

	 		if ( !strcmp(GetFieldNoStringMid(0,2,3),"511") || !strcmp(GetFieldNoStringMid(0,2,3),"512") )
		    	gbLateResultReturn=1; /*TRUE*/

				}
	
		if(PIPING && ulConnection) {
	    fprintf(pfOut,"%07lu",ulConnection);
			fprintf(pfOut,"%06lu",ulCurRecordNo);
			}
	
		
		for (uiCurFieldNo=0;uiCurFieldNo<uinFields;uiCurFieldNo++) {
                  /* ~ pstErrors= &Error_Object[uiCurFieldNo];		    */
			if(PIPING){
        uiValFieldResult=ValidateField(uiCurFieldNo,stErrors);
			  OutputCleanField(uiValFieldResult,uiCurFieldNo,pfOut);
	      }


			else if (!Options.bNoValid) {
				ValidateField(uiCurFieldNo,stErrors);
				}

			}
    
    
    
    if (!Options.bQuiet) {
    	/* Won't happen if piping */
			if (uiOutput)
				fprintf(stdout,"\n");
			          
		
			if (!((ulCurRecordNo-1)%50L)) {
				fprintf(stdout,"\n%08lu: ",ulCurRecordNo-1);
				}
					
			if (uinErrorStart!=uinError) {
				/* if (!Options.bRawOutput){
				   	fprintf(pfError,"\n"); 
 					} */    
				fprintf(stdout,"X");   
				} 
			else			
				fprintf(stdout,".");   
				}				
		
						  
		ulCurRecordNo++;			
		
		if (uinError>ulMaxErrors && ulMaxErrors && !PIPING) {
			fprintf(stdout,"\nThe maximum number of errors has been exceeded.\nProcessing will be terminated.\n");	
			}
		for (iC=0;iC<LBSIZE;iC++) {
			*(szLineBuffer+iC)=0;
			}


     if(PIPING) {
	    	fprintf(pfOut,"\n");
				}
		}
		
		

	ulCurRecordNo--;	
			                                
	if (!PIPING) {
		/* Close the input file */    
		fclose(pfIn);
		fclose(pfError);                                      
		}
	else {
		DoError1and4Report(stErrors ,stFields);
	
		fclose(pfIn);
		fclose(pfOut);
		
	    fclose(pfError1);
	    fclose(pfError2);
	    fclose(pfError3);
	}

			
    ulEnd=time(NULL);	                   
  
	UpDateInfoFile(argc,argv);

	/* Tidy up */
	                         
  if ((Options.bExplain) && prlRules) {
  	free((void *)prlRules);            
  	if (pfErrors)
  		fclose(pfErrors);
  	}
	
	/* 2004-03-30 (MDE) -b switch should make it pause*/
  	if (Options.bBreak)
		getchar();

	/* added 8 July to return corrct code to Ardvark*/
	if(uinError==0)
		exit(1);
		else
		exit(0);					
}             

